#include "framecpp/config.h"

#include <stdexcept>

#include "general/types.hh"
#include "general/autoarray.hh"
#include "general/reverse.hh"

#include "framecpp/Common/CompressionZeroSuppress.hh"

static const INT_2S MIN_INT_2S(
#if SIZEOF_SHORT == 2
			       SHRT_MIN
#else
#error Cannot compute the value for MIN_INT_2S
#endif
			       );
static const INT_4S MIN_INT_4S(
#if SIZEOF_INT == 4
			       INT_MIN
#else
#error Cannot compute the value for MIN_INT_4S
#endif
			       );

using General::AutoArray;

// To initially generate the log table algorithmically:
static const char* init_nBitTable256( );
template< typename T > INT_2U calcNBits( T Number );

static const char* nBitTable256( init_nBitTable256( ) );

template< > inline INT_2U
calcNBits( INT_2U Number )
{
  register INT_2U t;	// temporaries

  return ( ( t = ( Number >> 8 ) ) )
    ? 8 + nBitTable256[t]
    : nBitTable256[ Number ];
}

template< > inline INT_2U
calcNBits( INT_2S Number )
{
  return calcNBits( (INT_2U)Number );
}

template< > INT_2U
calcNBits( INT_4U Number )
{
  INT_2U r;     		// r will be lg(v)
  register INT_4U t, tt;	// temporaries

  if ( ( tt = ( Number >> 16) ) )
  {
    r = ( ( t = ( tt >> 8 ) ) )
      ? 24 + nBitTable256[t]
      : 16 + nBitTable256[tt];
  }
  else 
  {
    r = ( ( t = ( Number >> 8 ) ) )
      ? 8 + nBitTable256[t]
      : nBitTable256[ Number ];
  }
  return r;
}

template< > inline INT_2U
calcNBits( INT_4S Number )
{
  return calcNBits( (INT_4U)Number );
}

namespace
{
  static INT_4S ZERO_COMPRESS_SHORT_BSIZE = 12;
  static INT_4S ZERO_COMPRESS_INT_BSIZE = 8;

  void FrVectZExpandShort( INT_2S *out, const INT_2U *data, INT_8U nData );
  int  FrVectZCompShort ( INT_2U *out,
			  INT_8U *compL, const INT_2S *data, INT_8U nData,
			  INT_8U bSize = ZERO_COMPRESS_SHORT_BSIZE );

  void FrVectZExpandInt( INT_4S *out, const INT_4U *data, INT_8U nData );
  int  FrVectZCompInt ( INT_4U *out,
			INT_8U *compL, const INT_4S *data, INT_8U nData,
			INT_8U bSize = ZERO_COMPRESS_INT_BSIZE );

  template< typename InType > INT_8U
  zcomp_default_bsize( );

  template< > inline INT_8U
  zcomp_default_bsize< INT_2S >( )
  {
    return ZERO_COMPRESS_SHORT_BSIZE;
  }

  template< > inline INT_8U
  zcomp_default_bsize< INT_4S >( )
  {
    return ZERO_COMPRESS_INT_BSIZE;
  }

  template< typename InType, typename OutType > int
  zcomp( OutType *out,
	 INT_8U *compL, const InType *data, INT_8U nData,
	 INT_8U bSize );

  template < > inline int
  zcomp< INT_2S, INT_2U >( INT_2U *out,
			   INT_8U *compL, const INT_2S *data, INT_8U nData,
			   INT_8U bSize )
  {
    return FrVectZCompShort( out, compL, data, nData, bSize );
  }

  template < > inline int
  zcomp< INT_4S, INT_4U >( INT_4U *out,
			   INT_8U *compL, const INT_4S *data, INT_8U nData,
			   INT_8U bSize )
  {
    return FrVectZCompInt( out, compL, data, nData, bSize );
  }

  template< typename InType, typename OutType > void
  compress( const CHAR_U* Source,
	    INT_8U SourceLength,
	    General::AutoArray< CHAR_U >& Dest,
	    INT_8U& DestByteLength )
  {
    using namespace FrameCPP;

    try
    {
      typedef OutType out_type;
      typedef InType in_type;

      DestByteLength = 0;
      Dest.reset( (CHAR_U*)NULL );

      AutoArray< out_type >	comp_buffer( new out_type[ SourceLength ] );
      INT_8U			comp_buffer_len( SourceLength
						 * sizeof( out_type ) );

      zcomp( comp_buffer.get( ),
	     &comp_buffer_len,
	     reinterpret_cast< const in_type* >( Source ),
	     SourceLength,
	     zcomp_default_bsize< in_type >( ) );

      INT_4U ncdata( comp_buffer_len / sizeof( out_type ) );
      Dest.reset( new CHAR_U[ comp_buffer_len ] );
      std::copy( &comp_buffer[ 0 ], &comp_buffer[ ncdata ],
		 reinterpret_cast< out_type* >( Dest.get( ) ) );
      DestByteLength = comp_buffer_len;
    }
    catch( const FrUncompressable& Error )
    {
      // Return uncompressed buffer
    }
  }

} // namespace - anonymous

namespace FrameCPP
{
  namespace Compression
  {
    //-------------------------------------------------------------------
    /// The routines defined in this namespace are specific to the
    /// ZeroSuppress compression algorithm.
    /// Routines are defined for both compression and decompression
    /// of data buffers.
    //-------------------------------------------------------------------
    namespace ZeroSuppress
    {
      template< typename InType, typename OutType > void
      diff_compress( const CHAR_U* Source,
		     INT_8U SourceLength,
		     INT_4U DataType,
		     General::AutoArray< CHAR_U >& Dest,
		     INT_8U& DestByteLength )
      {
	try
	{
	  typedef OutType out_type;
	  typedef InType in_type;

	  DestByteLength = 0;
	  Dest.reset( (CHAR_U*)NULL );

	  AutoArray< CHAR_U >	diff_buffer;
	  AutoArray< out_type >	comp_buffer( new out_type[ SourceLength + 4 ] );
	  INT_8U		comp_buffer_len( ( SourceLength + 4 )
						 * sizeof( out_type ) );

	  Differential::Encode( Source, SourceLength, DataType, diff_buffer );
	  int ret = zcomp( comp_buffer.get( ), &comp_buffer_len,
			   reinterpret_cast< in_type* >( diff_buffer.get( ) ),
			   SourceLength,
			   zcomp_default_bsize< in_type >( ) );

	  if ( ret == 0 )
	  {
	    INT_4U ncdata( comp_buffer_len / sizeof( out_type ) );
	    Dest.reset( new CHAR_U[ comp_buffer_len ] );
	    std::copy( comp_buffer.get( ), comp_buffer.get( ) + ncdata,
		       reinterpret_cast< out_type* >( Dest.get( ) ) );
	    DestByteLength = comp_buffer_len;
	  }
	}
	catch( const FrUncompressable& Error )
	{
	  // Return uncompressed buffer
	}
      }

      template< typename InType, typename OutType > void
      expand( 	OutType* DataOut, const InType* DataIn, INT_8U NDataOut );

      template< > inline void
      expand< INT_2U, INT_2S >( INT_2S* DataOut, const INT_2U* DataIn,
				INT_8U NDataOut )
      {
	FrVectZExpandShort( DataOut, DataIn, NDataOut );
      }

      template< > inline void
      expand< INT_4U, INT_4S >( INT_4S* DataOut, const INT_4U* DataIn,
				INT_8U NDataOut )
      {
	FrVectZExpandInt( DataOut, DataIn, NDataOut );
      }

      template< typename InType, typename OutType > void
      expand( const CHAR_U* DataIn, INT_8U DataInBytes,
	      bool ByteSwap,
	      General::AutoArray< CHAR_U >& DataOut, INT_8U NDataOut,
	      INT_8U& NBytesOut )
      {
	typedef OutType out_type;
	typedef InType in_type;
	
	const INT_8U		data_out_byte_len=
	  ( NDataOut * sizeof( out_type ) );
	AutoArray< CHAR_U >	data_in_swap;

	const CHAR_U*		di( DataIn );

	if ( ByteSwap )
	{
	  //-------------------------------------------------------------
	  // Need a copy of the input buffer to put the bytes in the
	  // correct order
	  //-------------------------------------------------------------
	  data_in_swap.reset( new CHAR_U[ DataInBytes ] );
	  std::copy( DataIn, DataIn + DataInBytes,
		     data_in_swap.get( ) );
	  reverse< sizeof( in_type ) >( data_in_swap.get( ),
					DataInBytes / sizeof( in_type ) );
	  di = data_in_swap.get( );
	}
	AutoArray< CHAR_U >	data_out( new CHAR_U[ data_out_byte_len ] );

	expand( reinterpret_cast< out_type* >( data_out.get( ) ),
		reinterpret_cast< const in_type* >( di ),
		NDataOut );

	DataOut = data_out;
	NBytesOut = data_out_byte_len;
      }

      template<> void
      Compress< 2 >( const CHAR_U* Source,
		     INT_8U SourceLength,
		     General::AutoArray< CHAR_U >& Dest,
		     INT_8U& DestByteLength )
      {
	typedef INT_2U out_type;
	typedef INT_2S in_type;

	compress< in_type, out_type>( Source, SourceLength,
				      Dest, DestByteLength );
      }

      template<> void
      Compress< 4 >( const CHAR_U* Source,
		     INT_8U SourceLength,
		     General::AutoArray< CHAR_U >& Dest,
		     INT_8U& DestByteLength )
      {
	typedef INT_4U out_type;
	typedef INT_4S in_type;

	compress< in_type, out_type>( Source, SourceLength,
				      Dest, DestByteLength );
      }

      template< > void
      Expand< 2 >( const CHAR_U* DataIn, INT_8U DataInBytes,
		   bool ByteSwap,
		   General::AutoArray< CHAR_U >& DataOut, INT_8U NDataOut,
		   INT_8U& NBytesOut )
      {
	typedef INT_2S out_type;
	typedef INT_2U in_type;
	
	expand< in_type, out_type >( DataIn, DataInBytes, ByteSwap,
				     DataOut, NDataOut, NBytesOut );
      } // function template void Expand<2>()

      template< > void
      Expand< 4 >( const CHAR_U* DataIn, INT_8U DataInBytes,
		   bool ByteSwap,
		   General::AutoArray< CHAR_U >& DataOut, INT_8U NDataOut,
		   INT_8U& NBytesOut )
      {
	typedef INT_4S out_type;
	typedef INT_4U in_type;
	
	expand< in_type, out_type >( DataIn, DataInBytes, ByteSwap,
				     DataOut, NDataOut, NBytesOut );
      } // function template void xpand<4>()

      template<> void
      DiffCompress< 2 >( const CHAR_U* Source,
			 INT_8U SourceLength,
			 INT_4U DataType,
			 General::AutoArray< CHAR_U >& Dest,
			 INT_8U& DestByteLength )
      {
	typedef INT_2U out_type;
	typedef INT_2S in_type;

	diff_compress< in_type, out_type >( Source, SourceLength, DataType,
					    Dest, DestByteLength );
      }

      template<> void
      DiffCompress< 4 >( const CHAR_U* Source,
			 INT_8U SourceLength,
			 INT_4U DataType,
			 General::AutoArray< CHAR_U >& Dest,
			 INT_8U& DestByteLength )
      {
	typedef INT_4U out_type;
	typedef INT_4S in_type;

	diff_compress< in_type, out_type >( Source, SourceLength, DataType,
					    Dest, DestByteLength );
      }

      template < int Size > void
      DiffExpand( const CHAR_U* DataIn, INT_8U DataInBytes,
		  bool ByteSwap, INT_4U DataType,
		  General::AutoArray< CHAR_U >& DataOut, INT_8U NDataOut,
		  INT_8U& NBytesOut )
      {
	Expand< Size >( DataIn, DataInBytes, ByteSwap,
			DataOut, NDataOut, NBytesOut );

	if ( NBytesOut )
	{
	  Differential::Decode( DataOut.get( ), NDataOut,
				(data_types_type)DataType );
	}
      }

#define INSTANTIATE( LM_SIZE ) \
      template void \
      DiffExpand< LM_SIZE >( const CHAR_U* DataIn, INT_8U DataInBytes, \
			     bool ByteSwap, INT_4U DataType, \
			     General::AutoArray< CHAR_U >& DataOut, \
			     INT_8U NDataOut, INT_8U& NBytesOut )

      INSTANTIATE( 2 );
      INSTANTIATE( 4 );

#undef INSTANTIATE

    } // namespace - ZeroSuppress

  } // namespace - Compression

} // namespace - FrameCPP

namespace
{
  using FrameCPP::FrUncompressable;
  using FrameCPP::FrZCompRangeError;
  using FrameCPP::FrZCompIRangeError;

  template< typename wMaxType > const wMaxType*
  wMaxTable( )
  {
    static wMaxType table[ sizeof( wMaxType ) * 8 + 1 ];
    static bool initialized = false;

    if ( !initialized )
    {

      table[ 0 ] = 0;
      wMaxType	m1( 0x0 ), m2( 0x1 ), m3( 0x3 ), m4( 0x7 );

      INT_2U x = 0;
      for ( INT_2U c = 0 ; c < ( sizeof( wMaxType ) * 2 ); ++c )
      {
	table[ ++x ] = m1;
	table[ ++x ] = m2;
	table[ ++x ] = m3;
	table[ ++x ] = m4;
	m1 = ( m1 << 4 ) | 0xF;
	m2 = ( m2 << 4 ) | 0xF;
	m3 = ( m3 << 4 ) | 0xF;
	m4 = ( m4 << 4 ) | 0xF;
      }
      initialized = true;
    }

    return table;
  }

  template< typename maskType > const maskType*
  maskTable( )
  {
    static maskType table[ sizeof( maskType ) * 8 + 1];
    static bool initialized = false;

    if ( !initialized )
    {
      maskType	m1( 0x0 ), m2( 0x1 ), m3( 0x3 ), m4( 0x7 );

      INT_2U x = 0;
      for ( INT_2U c = 0 ; c < ( sizeof( maskType ) * 2 ); ++c )
      {
	table[ x++ ] = m1;
	table[ x++ ] = m2;
	table[ x++ ] = m3;
	table[ x++ ] = m4;
	m1 = ( m1 << 4 ) | 0xF;
	m2 = ( m2 << 4 ) | 0xF;
	m3 = ( m3 << 4 ) | 0xF;
	m4 = ( m4 << 4 ) | 0xF;
      }
      m1 = ( m1 << 4 ) | 0xF;
      table[ x ] = m1;
      initialized = true;
    }

    return table;
  }

  int FrVectZCompShort( INT_2U *out, 
			INT_8U *compL,
			const INT_2S *data,
			INT_8U nData,
			INT_8U bSize)
  {
    INT_2U nBits, uData, limit;
    INT_2S pOut;
    INT_2S max;
    INT_8U iIn, iOut, i, maxOut;

    static const INT_2S* wMax( wMaxTable< INT_2S >( ) );
    
    maxOut = *compL/2;
    /*------------------------- store the bloc size -----*/
    out[0] = bSize;
    iOut = 0;
    pOut = 16;
    /*--------------------------------- store the data --*/
    iIn = 0;
    while( iIn < nData )
    {
      /*------------- tune the size of the last bloc -----*/
      if(iIn + bSize > nData)
      {
	bSize = nData - iIn;
      }

      /*--------- get the maximum amplitude --------------*/
      
      max = 0;
      for( i=0; i<bSize; i++ )
      {
	if(data[iIn+i] == MIN_INT_2S )
	{
	  throw FrZCompRangeError( MIN_INT_2S );
	}
	if(data[iIn+i] > 0)
	{
	  max = max |  data[iIn + i];
	}
	else
	{
	  max = max | -data[iIn + i];
	}
      }

      /*--- determine the number of bits needed -( (2*(max-1)  < 2**nBits)) */

      nBits = calcNBits( max );
      
      /*---- encode the data size - we store nBits - 1 in 4 bits---*/

      out[iOut] = out[iOut] | ((nBits-1) << pOut);
      if(pOut > 12)
      {
	iOut ++;
	if (iOut >= maxOut)
	{
#if LM_INFO
	  std::cerr << "INFO: uncompressable: "
		    << " iOut: " << iOut
		    << " maxOut: " << maxOut
		    << std::endl;
#endif
	  throw FrUncompressable( "FrVectZCompShort: iOut >= maxOut" );
	}
	pOut = pOut - 16;
	out[iOut] = (nBits-1) >> - pOut;
      }
      pOut  = pOut +4;
      
      /*----------------------- encode the data itself ------*/
      if(nBits > 1)
      {
	limit = 16 - nBits;
	for(i = 0; i<bSize; i++)
	{
	  uData = (INT_2U) (data[iIn+i] + wMax[nBits]);
	  
	  out[iOut] = out[iOut] | (uData << pOut);
	  if(pOut > limit)
	  {
	    iOut ++;
	    if (iOut >= maxOut)
	    {
#if LM_INFO
	      std::cerr << "INFO: uncompressable: "
			<< " iOut: " << iOut
			<< " maxOut: " << maxOut
			<< std::endl;
#endif
	      throw FrUncompressable( "FrVectZCompShort: iOut >= maxOut" );
	    }
	    pOut = pOut - 16;
	    out[iOut] = uData >> -pOut;}
	  pOut  = pOut + nBits;
	}
      } 
      /*----------------------------- increase pointer -------*/
      
      iIn = iIn + bSize;
    }

    *compL = 2*(iOut+1);

    return(0);
  }

  int
  FrVectZCompInt( INT_4U*	out, 
		  INT_8U*	compL,
		  const INT_4S*	data,
		  INT_8U	nData,
		  INT_8U	bSize )
  {
    INT_4U nBits, uData, limit;
    INT_4U pOut;
    INT_4S max;
    INT_8U  iIn, iOut, i, maxOut;
    static const INT_4S* wMax( wMaxTable< INT_4S >( ) );

    maxOut = *compL/4;
    /*----------------------------store the bloc size--*/
    out[0] = bSize;
    iOut = 0;
    pOut = 16;
    /*------------------------------- store the data --*/
    iIn = 0;
    while(iIn<nData)
    {
      /*------------ tune the size of the last bloc -----*/

      if(iIn + bSize > nData)
      {
	bSize = nData - iIn;
      }

      /*-------- get the maximum amplitude --------------*/
      max = 0;
     
      for( i=0; i < bSize; i++)
      {
	if( data[iIn+i] == MIN_INT_4S )
	{
#if LM_INFO
	  std::cerr << "INFO: iIn+i: " << ( iIn+i )
		    << " " << iIn << " " << i
		    << " data: " << data[iIn+i]
		    << std::endl;
#endif
	  throw FrameCPP::FrZCompIRangeError( MIN_INT_4S );
	}
	if(data[iIn+i] > 0)
	{
	  max = max |  data[iIn + i];
	}
	else
	{
	  max = max | -data[iIn + i];
	}
      }

      /*----- determine the number of bits needed - (2*(max-1)  < 2**nBits)   */
    
      nBits = calcNBits< INT_4U >( max );
      
      /*---- encode the data size ---- we store nBits - 1 in 4 bits---*/

      if(pOut != 32)
      {
	out[iOut] = out[iOut] | ((nBits-1) << pOut);
      }
      if(pOut > 27)
      {
	iOut ++;
	if(iOut >= maxOut)
	{
#if LM_INFO
	  std::cerr << "INFO: uncompressable: "
		    << " iOut: " << iOut
		    << " maxOut: " << maxOut
		    << std::endl;
#endif
	  throw FrUncompressable( "FrVectZCompInt: iOut >= maxOut" );
	}
	pOut = pOut - 32;
	out[iOut] = (nBits-1) >> - pOut;}
      pOut  = pOut + 5;
      /*--------- encode the data itself ------*/
      if(nBits > 1)
      {
	limit = 32 - nBits;
	for(i = 0; i<bSize; i++)
	{
	  uData = (unsigned int) (data[iIn+i] + wMax[nBits]);
	  
	  if( pOut != 32) out[iOut] = out[iOut] | (uData << pOut);
	  if( pOut > limit)
	  {
	    iOut ++;
	    if(iOut >= maxOut)
   	    {
#if LM_INFO
	      std::cerr << "INFO: uncompressable: "
			<< " iOut: " << iOut
			<< " maxOut: " << maxOut
			<< std::endl;
#endif
	      throw FrUncompressable( "FrVectZCompInt: iOut >= maxOut" );
	    }
	    pOut = pOut - 32;
	    out[iOut] = uData >> -pOut;
	  }
	  pOut  = pOut + nBits;
	}
      } 
      /*------------------------- increase pointer -------*/
  
      iIn = iIn + bSize;
    }

    *compL = 4*(iOut+1);

    return(0);
  }

  void
  FrVectZExpandShort( INT_2S*		out,
		      const INT_2U*	data, 
		      INT_8U		nData)
  {
    static const INT_2S* wMax( wMaxTable< INT_2S >( ) );
    static const INT_2U* mask( maskTable< INT_2U >( ) );

    union {
      INT_4U	int_4u;
      INT_2U	int_2u[2];
    } iBuf;

    INT_2U nBits, pIn, uData, *buf;
    INT_8U i, iIn, iOut, bSize;

    /*------------------ retrieve the bloc size -----*/
    bSize = data[0];
    iIn = 1;
    pIn = 0;
    /*------------- retrieve the data  --------------*/

    buf = (INT_2U *) &(iBuf.int_2u[0]);
    iOut = 0;
    do
    {
      /*-------- extract nBits ---(we check if data is in 1 or 2 words) ------*/
        
      if(pIn <= 12)
      {
	uData = data[iIn] >> pIn;
	pIn = pIn + 4;
      }
      else
      {
	buf[0] = data[iIn];
	iIn++;
#if WORDS_BIGENDIAN
	buf[1] = buf[0];
	buf[0] = data[iIn];
#else /* WORDS_BIGENDIAN */
	buf[1] = data[iIn];
#endif /* WORDS_BIGENDIAN */
	uData = iBuf.int_4u >> pIn;
	pIn = pIn - 12;
      }
      
      nBits = 1 + (mask[4] & uData);
      if(nBits == 1)
	nBits = 0;

      /*----------------------------extract data ----------*/
    
      for(i=0; i<bSize; i++)
      {
	if(iOut >= nData)
	{
	  break;
	}

	if(pIn + nBits <= 16)
	{
	  uData = data[iIn] >> pIn;
	  pIn = pIn + nBits;
	}
	else
	{
	  buf[0] = data[iIn];
	  iIn++;
#if WORDS_BIGENDIAN
	  buf[1] = buf[0];
	  buf[0] = data[iIn];
#else /* WORDS_BIGENDIAN */
	  buf[1] = data[iIn];
#endif /* WORDS_BIGENDIAN */
	  uData = iBuf.int_4u >> pIn;
	  pIn = pIn + nBits - 16;
	}

	out[iOut] = (mask[nBits] & uData) - wMax[nBits];
	iOut++;
      }
    }
    while(iOut<nData);
  }

  void
  FrVectZExpandInt( INT_4S*		out, 
		    const INT_4U*	data, 
		    INT_8U		nData )
  {
    typedef INT_4S out_type;
    typedef INT_4U in_type;

    INT_4U nBits, pIn, uData;
    static const in_type* wMax( wMaxTable< in_type >( ) );
    static const in_type* mask( maskTable< in_type >( ) );

    INT_4S  i, bSize;
    INT_4U iOut, iIn;

    /*---------------- retrieve the bloc size -----*/
    bSize = data[0] & 0xffff;

    iIn = 0;
    pIn = 16;

    /*------------------ retrieve the data---------*/

    iOut = 0;
    do
    {
      /*-------- extract nBits -(we check if data is in 1 or 2 words) ------*/
        
      if(pIn <= 27)
      {
	uData = data[iIn] >> pIn;
	pIn = pIn + 5;
      }
      else
      {
	uData = (data[iIn] >> pIn) & mask[32-pIn];
	iIn++;
	uData += data[iIn] << (32-pIn);
	pIn = pIn - 27;
      }

      nBits = 1 + (0x1f & uData);
      if(nBits == 1)
      {
	nBits = 0;
      }

      /*----------------------extract data ----------*/
      for(i=0; i<bSize; i++)
      {
	if(iOut >= nData)
	{
	  break;
	}
	
	if(pIn + nBits <= 32)
	{
	  uData = data[iIn] >> pIn;
	  pIn = pIn + nBits;
	}
	else
	{
	  uData = (data[iIn] >> pIn) & mask[32-pIn];
	  iIn++;
	  uData += data[iIn] << (32-pIn);
	  pIn = pIn + nBits - 32;
	}

	out[iOut] = (mask[nBits] & uData) - wMax[nBits];
	iOut++;
      }
    }
    while(iOut<nData);
  }

}

static const char*
init_nBitTable256( )
{
  static char table[256 ];

  table[0] = 1;
  table[1] = 2;

  for (int i = 2; i < 256; i++) 
  {
    table[i] = 1 + table[i / 2];
  }
  return table;
}
